package gearman

import (
	"container/list"
	"sync"
)

// SyncMessageFIFO provides a synced FIFO list for gearman messages
// First in First out (FIFO)
// The default Go lang channels are blocked so in case that a Satellite System is not reachable
// The Go lang channel would block forever.
// The FIFO "channel" the "input" channel to put results into a cache. When the cache has items, the "notify" channel get's signaled
// If the Satellite System is not reachable, the messages will be added back to the top of the cache via "inputFront"
type SyncMessageFIFO struct {
	noCopy noCopy // do not copy by value

	mtx      sync.Mutex
	messages list.List

	input      chan *Message
	inputBack  chan []*Message
	inputFront chan []*Message
	notify     chan struct{}

	closer chan struct{}
	wg     sync.WaitGroup
}

func (f *SyncMessageFIFO) run() {
	defer func() {
		close(f.input)
		close(f.inputFront)
		close(f.inputBack)
		close(f.notify)
		close(f.closer)
		f.wg.Done()
	}()

	// Create a dummy channel
	noNotify := make(chan struct{})

	for {
		f.mtx.Lock()
		notify := noNotify
		if f.messages.Len() > 0 {
			// We have messages in cache - notify the real channel of the receiver
			notify = f.notify
		}
		f.mtx.Unlock()
		select {
		case <-f.closer:
			return
		case notify <- struct{}{}: // send notify signal to receiver
		case msg := <-f.input:
			// Got message on input channel and push it to the end of the cache
			f.mtx.Lock()
			f.messages.PushBack(msg)
			f.mtx.Unlock()
		case msgs := <-f.inputFront:
			// Got message on inputFront channel and push it to the top of the cache
			f.mtx.Lock()
			for _, msg := range msgs {
				f.messages.PushFront(msg)
			}
			f.mtx.Unlock()
		case msgs := <-f.inputBack:
			// Same as input but supports an array of messages
			f.mtx.Lock()
			for _, msg := range msgs {
				f.messages.PushBack(msg)
			}
			f.mtx.Unlock()
		}
	}
}

// Start creates a background process for io i/o
func (f *SyncMessageFIFO) Start() {
	f.input = make(chan *Message)
	f.inputFront = make(chan []*Message)
	f.inputBack = make(chan []*Message)
	f.notify = make(chan struct{})
	f.closer = make(chan struct{})
	f.wg.Add(1)
	go f.run()
}

// Close sends a signal to the gopher to stop
func (f *SyncMessageFIFO) Close() {
	f.closer <- struct{}{}
	f.wg.Wait()
}

// Notify handles a notification if a new message is available in the buffer
func (f *SyncMessageFIFO) Notify() chan struct{} {
	return f.notify
}

// Push adds a message to the FIFO
func (f *SyncMessageFIFO) Push(msg *Message) {
	f.input <- msg
}

// PushAll adds a slice of messages to the FIFO
func (f *SyncMessageFIFO) PushAll(msgs []*Message, front bool) {
	if front {
		f.inputFront <- msgs
	} else {
		f.inputBack <- msgs
	}
}

// Get returns one elemnt from the FIFO
func (f *SyncMessageFIFO) Get() *Message {
	f.mtx.Lock()
	defer f.mtx.Unlock()

	el := f.messages.Front()
	if el == nil {
		return nil
	}
	msg := el.Value.(*Message)

	f.messages.Remove(el)
	return msg
}

// GetAll returns up to max messages from the FIFO
func (f *SyncMessageFIFO) GetAll(max int) []*Message {
	f.mtx.Lock()
	defer f.mtx.Unlock()

	lenMsg := f.messages.Len()
	if lenMsg > max {
		lenMsg = max
	}
	msgs := make([]*Message, lenMsg)

	for i := 0; i < lenMsg; i++ {
		el := f.messages.Front()
		if el == nil {
			msgs = msgs[:i]
			break
		}
		msgs[i] = f.messages.Remove(el).(*Message)
	}

	return msgs
}

// Direct returns the channel were you can push messages directly into the buffer
func (f *SyncMessageFIFO) Direct() chan<- *Message {
	return f.input
}
